import importlib
import json
from collections import OrderedDict
from copy import deepcopy
from hashlib import md5
import warnings
import itertools

import inflection
from dateutil.parser import parse as parse_dt


def hash_dict(dictionary):
    return md5(json.dumps(dictionary, sort_keys=True).encode("utf-8")).hexdigest()


FILTERS = {
    #'exact',
    #'iexact',
    #'contains',
    #'icontains',
    "in": lambda x, y: x in y,
    "eq": lambda x, y: x == y,
    #'gt',
    #'gte',
    #'lt',
    #'lte',
    #'startswith',
    #'istartswith',
    #'endswith',
    #'iendswith',
    #'range',
    #'date',
    #'year','
    #'month',
    #'day',
    #'week',
    #'week_day',
    #'quarter','
    #'time',
    #'hour','
    #'minute',
    #'second',
    #'isnull',
    #'regex',
    #'iregex',
}


def one_to_one(class_name, **mapping):
    module_name, class_name = class_name.rsplit(".", 1)

    @property
    def get(self):
        try:
            klass = getattr(importlib.import_module(module_name), class_name)
        except AttributeError:
            return None
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.get(**filter_obj)

    return get


def one_to_many(class_name, **mapping):
    module_name, class_name = class_name.rsplit(".", 1)

    @property
    def get(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.filter(**filter_obj)

    return get


def values_iterator(iterator, *keys, tuples=False):
    for v in iterator:
        if tuples:
            yield tuple(recur_accessor(v, k) for k in keys)
        else:
            yield {k: recur_accessor(v, k) for k in keys}


def recur_accessor(obj, accessor):
    if "__" not in accessor:
        a = accessor
        rest = None
    else:
        a, rest = accessor.split("__", 1)

    if hasattr(obj, a):
        new_obj = getattr(obj, a)
        if callable(new_obj):
            new_obj = new_obj()
    else:
        try:
            a = int(a)
        except ValueError:
            pass
        try:
            new_obj = obj[a]
        except (KeyError, TypeError, IndexError):
            new_obj = None
    if not rest:
        return new_obj
    else:
        return recur_accessor(new_obj, rest)


class Model(object):
    """
    Model Base Class

    Takes in a dictionary of data and :
    1. Automatically inflects all keys to underscore_case
    2. Converts any value that has "date" or "datetime" into the appropriate type
    3. Appends all keys in the dictionary as attributes on the object
    4. Attaches the original inflected data dictionary as Model.data
    """

    def __init__(self, data, backref=None, **kwargs):
        try:
            self.data = self._convert_data(data)

            for k, v in self.data.items():
                setattr(self, k, self.data[k])
        except Exception as e:
            self.error = e
        self.backref = backref

    def _convert_data(self, data):
        if isinstance(data, list):
            return [self._convert_data(a) for a in data]
        elif isinstance(data, dict):
            dictionary = {
                inflection.underscore(k): self._convert_data(v)
                for (k, v) in data.items()
            }
            for k, v in dictionary.items():
                try:
                    if "datetime" in k and type(v) == str:
                        dictionary[k] = parse_dt(v)
                    elif "date" in k and type(v) == str:
                        dictionary[k] = parse_dt(v).date()
                except ValueError:  # Malformed datetimes:
                    dictionary[k] = None
            return dictionary
        else:
            return data

    def as_dict(self):
        """Convert object to dictionary, recursively converting any objects to dictionaries themselves"""
        output = {key: getattr(self, key) for key in self.attrs if hasattr(self, key)}
        for k, v in output.items():
            if isinstance(v, list):
                output[k] = [i.asdict() if hasattr(i, "asdict") else i for i in v]
            if hasattr(v, "asdict"):
                output[k] = v.asdict()
        return output

    def __repr__(self):
        """Default representation"""
        primary_key = getattr(self, "primary_key", self.attrs[0])
        repr_string = (
            f"<{self.__class__.__name__} {primary_key}={getattr(self, primary_key)}"
        )
        if hasattr(self, "error"):
            repr_string += f" error={self.error}"
        return repr_string + ">"

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return (self.__class__ == other.__class__) and (hash(self) == hash(other))

    def __ne__(self, other):
        return not self.__eq__(other)


class BaseManager:
    """
    Manager Base Class

    This class is essentially a configurable generator. It is intended to be initialized 
    as an empty object at Model.objects. Users can then modify the manager by calling either:
        
        Manager.filter -> adds filter criteria
        Manager.sort -> adds sorting criteria
        Manager.options -> adds key:value options

    All methods should return a brand-new manager with the appropriate parameters re-set. The 
    Manager also has a special method to fetch a single matching object:

        Manager.get -> adds filter critera, and returns the first object if only one object is found, else raises an Exception

    The manager's attributes are stored in a dictionary at Manager.config. Config has the following structure:

        {
            'filter': {
                key: value,
                key: value,
            },
            'order_by': [value, value, value],
            'values': [value, value, value]
            'options': {
                key: value
            }
        }
        
    """

    def __init__(
        self, config=dict(filter=dict(), order_by=list(), options=dict()), values=list()
    ):
        self.config = config

    def __add__(self, other):
        return QuerySet(itertools.chain(self, other))

    def filter(self, *args, **kwargs):
        if args:
            kwargs[self.primary_key] = args
        for k, v in kwargs.items():
            if not isinstance(v, str):
                try:
                    kwargs[k] = list(str(i) for i in v)
                except TypeError:
                    kwargs[k] = str(v)

        new_config = deepcopy(self.config)
        new_config["filter"] = {**new_config["filter"], **kwargs}
        return self.__class__(new_config)

    def order_by(self, *args):
        """Take arguments, and store in a special keyword argument called 'sort' """
        new_config = deepcopy(self.config)
        new_config["order_by"] = list(new_config["order_by"]) + list(args)
        return self.__class__(new_config)

    def set_options(self, **kwargs):
        new_config = deepcopy(self.config)
        new_config["options"] = {**new_config["options"], **kwargs}
        return self.__class__(new_config)

    def get(self, *args, **kwargs):
        """Implement a new manager with the requested keywords, and if the length is 1,
        return that record, else raise an exception"""
        manager = self.filter(*args, **kwargs)
        if len(manager) > 1:
            doc_nos = "\n".join([str(r) for r in manager])
            raise ValueError("More than one document found!\n" + doc_nos)
        return manager.first()

    def values(self, *fields):
        """Return new manager with special keywords 'values__fields' and 'values__list'"""
        new_config = deepcopy(self.config)
        new_config["values"] = fields
        new_config["options"]["values_iterator"] = True
        new_config["options"]["values_list"] = False
        return self.__class__(new_config)

    def values_list(self, *fields, flat=False):
        """Same as values, but adds an additional parameter for "flat" lists """
        new_config = deepcopy(self.config)
        new_config["values"] = fields
        new_config["options"]["values_iterator"] = True
        new_config["options"]["values_list"] = True
        new_config["options"]["values_list_flat"] = flat
        return self.__class__(new_config)

    def to_pandas(self, limit=50):
        import pandas as pd

        objects = list(self)
        return pd.DataFrame.from_records(
            {k: getattr(r, k, None) for k in objects[0].attrs} for r in objects
        )

    def explode(self, attribute):
        from itertools import chain

        return QuerySet(chain.from_iterable(getattr(r, attribute, None) for r in self))


class QuerySet(BaseManager):
    """
    Utility class that extends the Manager helper function to 
    any collection of Patent Client objects
    """

    def __init__(self, iterable):
        self.iterable = iterable

    def __getitem__(self, key):
        return self.iterable[key]

    def __iter__(self):
        return iter(self.iterable)

    def __repr__(self):
        return f"<QuerySet({repr(self.iterable)})>"


class IterableManager(BaseManager):
    """
    Iterable Manager Base Class

    This class extends BaseManager for iterable managers. This requires that the extending manager
    implement __iter__. 

    This is used for all managers that are not container-like - i.e. does not provide a definite length

    """

    def first(self):
        return next(iter(self))


class Manager(BaseManager):
    """
    Standard Manager Base Class

    This class extends BaseManager by adding support for the __getitem__ method for iteration
    This is used for all managers that behave like containers, rather than pure iterators

    """

    def count(self):
        return len(self)

    def first(self):
        try:
            return self[0]
        except StopIteration:
            raise ValueError("No matching records found!")

    def all(self):
        return iter(self)

    def get_item(self, key):
        raise NotImplementedError(f"{self.__class__} has no get_item method")

    def __getitem__(self, key):
        """resolves slices and keys into Model objects. Relies on .get_item(key) to obtain
        the record itself"""
        if type(key) == slice:
            indices = list(range(len(self)))[key.start : key.stop : key.step]
            return [self.__getitem__(index) for index in indices]
        else:
            if key >= len(self):
                raise StopIteration
            obj = self.get_item(key)
            options = self.config["options"]

            if options.get("values_iterator") == True:
                data = obj.data
                fdata = OrderedDict()
                for k in self.config["values"]:
                    value = recur_accessor(obj, k)
                    fdata[k] = value
                data = fdata
                if options.get("values_list", False):
                    data = tuple(data[k] for k, v in data.items())
                    if len(self.config["values"]) == 1 and options.get(
                        "values_list_flat", False
                    ):
                        data = data[0]
                return data
            else:
                return obj



from marshmallow import EXCLUDE
from marshmallow import fields
from marshmallow import post_load
from marshmallow import pre_load
from marshmallow import Schema
from patent_client.util import ListField
from patent_client.util import QuerySet
from patent_client.util.manager import resolve

from .model import CpcClass
from .model import Inpadoc
from .model import InpadocApplication
from .model import InpadocBiblio
from .model import InpadocPriorityClaim
from .model import InpadocPublication


class BaseSchema(Schema):
    @post_load
    def make_object(self, data, **kwargs):
        if hasattr(self, "__model__"):
            return self.__model__(**data)
        return data


class InpadocResultSchema(BaseSchema):
    __model__ = Inpadoc
    country = fields.Str(data_key="country")
    number = fields.Str(data_key="doc-number")
    kind_code = fields.Str(data_key="kind")
    doc_type = fields.Str(data_key="@document-id-type")
    family_id = fields.Int(data_key="@family-id")

    @pre_load
    def pre_load(self, data, *args, **kwargs):
        for k, v in data["document-id"].items():
            data[k] = v
        del data["document-id"]
        return data

    class Meta:
        unknown = EXCLUDE


def flatten(data, subfield):
    data = {**data, **data[subfield]}
    del data[subfield]
    return data


class InpadocSchema(BaseSchema):
    __model__ = Inpadoc
    country = fields.Str(data_key="country")
    number = fields.Str(data_key="doc-number")
    kind_code = fields.Str(data_key="kind")
    doc_type = fields.Str(data_key="@document-id-type")
    date = fields.Date(format="%Y%m%d")


class InpadocApplicationSchema(InpadocSchema):
    __model__ = InpadocApplication

    class Meta:
        unknown = EXCLUDE


class InpadocPublicationSchema(InpadocSchema):
    __model__ = InpadocPublication

    class Meta:
        unknown = EXCLUDE


class PriorityClaimSchema(BaseSchema):
    __model__ = InpadocPriorityClaim
    doc_type = fields.Str(data_key="@document-id-type")
    kind_code = fields.Str(data_key="@kind")
    date = fields.Date(format="%Y%m%d")
    number = fields.Str(data_key="doc-number")

    class Meta:
        unknown = EXCLUDE


class CpcClassSchema(BaseSchema):
    __model__ = CpcClass
    sequence = fields.Str(data_key="@sequence")
    section = fields.Str()
    cpc_class = fields.Str(data_key="class")
    subclass = fields.Str()
    main_group = fields.Str(data_key="main-group")
    sub_group = fields.Str(data_key="subgroup")
    classification_value = fields.Str(data_key="classification-value")
    generating_office = fields.Str(data_key="generating-office")

    class Meta:
        unknown = EXCLUDE


class UsClassSchema(BaseSchema):
    symbol = fields.Str(data_key="classification-symbol")

    class Meta:
        unknown = EXCLUDE


class InpadocBiblioSchema(BaseSchema):
    __model__ = InpadocBiblio

    family_id = fields.Str(data_key="@family-id")
    country = fields.Str(data_key="@country")
    number = fields.Str(data_key="@doc-number")
    kind_code = fields.Str(data_key="@kind")
    abstract = fields.Method(deserialize="get_abstract")
    inventors = fields.Method(deserialize="get_inventors")
    applicants = fields.Method(deserialize="get_applicants")
    publications = ListField(fields.Nested(InpadocPublicationSchema))
    ipc_classes = ListField(fields.Str(), allow_none=True)
    cpc_classes = ListField(fields.Nested(CpcClassSchema), allow_none=True)
    us_classes = ListField(fields.Nested(UsClassSchema), allow_none=True)
    applications = ListField(fields.Nested(InpadocApplicationSchema))
    priority_claims = ListField(fields.Nested(PriorityClaimSchema))
    title = fields.Str()

    def get_abstract(self, data):
        if isinstance(data, dict):
            return data["p"]
        else:
            abstract = next((d for d in data if d["@lang"] == "en"), data[0])
            return abstract["p"]

    def get_inventors(self, data):
        return [
            i["inventor-name"]["name"] for i in data if i["@data-format"] == "original"
        ]

    def get_applicants(self, data):
        return [
            i["applicant-name"]["name"] for i in data if i["@data-format"] == "original"
        ]

    def pre_load_priority_claims(self, bib):
        pcs = resolve(bib, "priority-claims.priority-claim")
        out = list()
        pcs = (
            pcs if isinstance(pcs, list) else [pcs,]
        )
        for pc in pcs:
            if isinstance(pc["document-id"], list):
                doc_id = pc["document-id"][0]
            else:
                doc_id = pc["document-id"]
            pc = {**pc, **doc_id}
            del pc["document-id"]
            out.append(pc)
        return out

    @pre_load
    def pre_load(self, data, *args, **kwargs):
        data = data["exchange-document"]
        bib = data["bibliographic-data"]
        del data["bibliographic-data"]
        data["publications"] = resolve(bib, "publication-reference.document-id")
        ipc_class = resolve(bib, "classifications-ipcr.classification-ipcr")
        if ipc_class:
            data["ipc_classes"] = [c["text"] for c in ipc_class]
        classifications = resolve(bib, "patent-classifications.patent-classification")
        if classifications:
            data["cpc_classes"] = [
                c
                for c in classifications
                if resolve(c, "classification-scheme.@scheme") == "CPCI"
            ]
            data["us_classes"] = [
                c
                for c in classifications
                if resolve(c, "classification-scheme.@scheme") == "UC"
            ]
        data["applications"] = resolve(bib, "application-reference.document-id")
        data["priority_claims"] = self.pre_load_priority_claims(bib)
        data["applicants"] = resolve(bib, "parties.applicants.applicant")
        data["inventors"] = resolve(bib, "parties.inventors.inventor")
        titles = resolve(bib, "invention-title")
        if isinstance(titles, list):
            data["title"] = next(t["#text"] for t in titles if t["@lang"] == "en")
        elif isinstance(titles, str):
            data["title"] = titles
        else:
            data["title"] = titles["#text"]
        return data

    class Meta:
        unknown = EXCLUDE



import datetime
import logging
from collections import OrderedDict

import pytest

from .model import USApplication

logger = logging.getLogger(__name__)


class TestPatentExaminationData:
    def test_get_inventors(self):
        inventors = USApplication.objects.get(12721698).inventors
        assert len(inventors) == 1
        inventor = inventors[0]
        assert inventor.name == "Thind; Deepinder Singh"
        assert inventor.city == "Mankato, "
        assert inventor.geo_code == "MN"

    def test_search_by_customer_number(self):
        result = USApplication.objects.filter(app_cust_number="70155")
        assert len(result) > 1

    def test_get_by_pub_number(self):
        pub_no = "US20060127129A1"
        app = USApplication.objects.get(app_early_pub_number=pub_no)
        assert app.patent_title == "ELECTROPHOTOGRAPHIC IMAGE FORMING APPARATUS"

    def test_get_by_pat_number(self):
        pat_no = 6095661
        app = USApplication.objects.get(patent_number=pat_no)
        assert app.patent_title == "METHOD AND APPARATUS FOR AN L.E.D. FLASHLIGHT"

    def test_get_by_application_number(self):
        app_no = "15145443"
        app = USApplication.objects.get(app_no)
        assert (
            app.patent_title
            == "Suction and Discharge Lines for a Dual Hydraulic Fracturing Unit"
        )

    def test_get_many_by_application_number(self):
        app_nos = ["14971450", "15332765", "13441334", "15332709", "14542000"]
        data = USApplication.objects.filter(*app_nos)
        assert len(data) == 5

    def test_search_patex_by_assignee(self):
        data = (
            USApplication.objects.filter(first_named_applicant="LogicBlox")
            .order_by("appl_id")
            .limit(4)
        )
        expected_titles = [
            "MAINTENANCE OF ACTIVE DATABASE QUERIES",
            "SALIENT SAMPLING FOR QUERY SIZE ESTIMATION",
            "TRANSACTION REPAIR",
            "LEAPFROG TREE-JOIN",
        ]
        app_titles = data.values_list("patent_title", flat=True)
        app_titles = [a.upper() for a in app_titles]
        for t in expected_titles:
            assert t in app_titles

    def test_get_many_by_publication_number(self):
        nos = [
            "US20080034424A1",
            "US20100020700A1",
            "US20110225644A1",
            "US20050120054A1",
            "US20050188423A1",
        ]
        data = USApplication.objects.filter(app_early_pub_number=nos)
        assert len(list(data)) == 5

    def test_get_search_fields(self):
        result = USApplication.objects.allowed_filters
        assert "patent_number" in result
        assert "appl_id" in result
        assert "app_early_pub_number" in result

    def test_get_child_data(self):
        parent = USApplication.objects.get("14018930")
        child = parent.child_continuity[0]
        assert child.child_appl_id == "14919159"
        assert child.relationship == "claims the benefit of"
        # assert child.child.patent_title == "LEAPFROG TREE-JOIN"

    def test_get_parent_data(self):
        child = USApplication.objects.get("14018930")
        parent = child.parent_continuity[0]
        assert parent.parent_appl_id == "61706484"
        assert parent.relationship == "Claims Priority from Provisional Application"
        assert parent.parent_app_filing_date is not None
        # assert parent.parent.patent_title == "Leapfrog Tree-Join"

    def test_pta_history(self):
        app = USApplication.objects.get("14095073")
        pta_history = app.pta_pte_tran_history
        assert len(pta_history) > 10

    def test_pta_summary(self):
        app = USApplication.objects.get("14095073")
        expected = OrderedDict(
            [
                ("kind", "PTA"),
                ("a_delay", 169),
                ("pto_delay", 169),
                ("applicant_delay", 10),
                ("total_days", 159),
            ]
        )
        actual = app.pta_pte_summary.as_dict()
        for k, v in expected.items():
            assert actual[k] == v

    def test_transactions(self):
        app = USApplication.objects.get("14095073")
        assert len(app.transactions) > 70

    def test_correspondent(self):
        app = USApplication.objects.get("14095073")
        correspondent = app.correspondent.as_dict()
        expected = {
            "name": "VINSON & ELKINS L.L.P.",
            "cust_no": "22892",
            "street": "1001 Fannin Street\nSuite 2500",
            "city": "HOUSTON",
            "geo_region_code": "TX",
            "postal_code": "77002-6760",
        }
        for k in expected.keys():
            assert expected[k] == correspondent[k]

    def test_attorneys(self):
        app = USApplication.objects.get("14095073")
        assert len(app.attorneys) > 1
        actual = app.attorneys[0].as_dict()
        expected = {
            "registration_no": "32429",
            "full_name": "Mims, Peter  ",
            "phone_num": "713-758-2732",
            "reg_status": "ACTIVE",
        }
        for k in expected.keys():
            assert expected[k] == actual[k]

    def test_iterator(self):
        apps = USApplication.objects.filter(first_named_applicant="Tesla").limit(68)
        counter = 0
        try:
            for a in apps:
                counter += 1
        except KeyError as e:
            raise e
        assert len(apps) == counter

    def test_expiration_date(self):
        app = USApplication.objects.get("15384723")
        expected = {
            "parent_appl_id": "12322218",
            "parent_app_filing_date": datetime.date(2009, 1, 29),
            "parent_relationship": "is a Continuation in part of",
            "initial_term": datetime.date(2029, 1, 29),
            "pta_or_pte": 0,
            "extended_term": datetime.date(2029, 1, 29),
            "terminal_disclaimer_filed": True,
        }
        actual = dict(app.expiration)
        for k in expected.keys():
            assert expected[k] == actual[k]

        app = USApplication.objects.get("14865625")
        expected = {
            "parent_appl_id": "14865625",
            "parent_app_filing_date": datetime.date(2015, 9, 25),
            "parent_relationship": "self",
            "initial_term": datetime.date(2035, 9, 25),
            "pta_or_pte": 752,
            "extended_term": datetime.date(2037, 10, 16),
            "terminal_disclaimer_filed": False,
        }
        actual = dict(app.expiration)
        for k in expected.keys():
            assert expected[k] == actual[k]

    def test_expiration_date_for_pct_apps(self):
        app = USApplication.objects.get("PCT/CA02/01413")
        with pytest.raises(Exception) as exc:
            expiration_data = app.expiration
        assert exc.match("Expiration date not supported for PCT Applications")

    def test_issue_25(self):
        company_name = "Tesla"
        records = list(
            USApplication.objects.filter(first_named_applicant=company_name)
            .limit(5)
            .values("app_filing_date", "patent_number", "patent_title")[:]
        )
        assert len(records) >= 4

    def test_get_applicant(self):
        applicants = (
            USApplication.objects.filter(first_named_applicant="Tesla")
            .first()
            .applicants
        )
        assert "tesla" in applicants[0].name.lower()

    def test_foreign_priority(self):
        fp = USApplication.objects.get(patent_number=10544653).foreign_priority
        assert isinstance(fp, list)
        app = fp[0]
        assert app.country_name == "NORWAY"
        assert app.filing_date == datetime.date(2017, 2, 15)
        assert app.priority_claim == "20170229"



from pprint import pprint
from patent_client.epo.inpadoc.model import Inpadoc, InpadocBiblio


class TestInpadoc():
    def test_inpadoc_manager(self):
        result = Inpadoc.objects.filter(applicant="Microsoft")
        assert len(result) > 20
        countries = list(result.limit(20).values_list('country', flat=True))
        assert sum(1 for c in countries if c == 'US') >= 1
       
    def test_get_biblio_from_result(self):
        result = Inpadoc.objects.filter(applicant="Google").first().biblio
        assert result.title is not None

    def test_get_claims_from_result(self):
        result = Inpadoc.objects.get(publication="WO2009085664A2")
        assert len(result.claims) == 4833

    def test_get_description_from_result(self):
        result = Inpadoc.objects.get(publication="WO2009085664A2")
        assert len(result.description) == 47956

    def test_get_family_from_result(self):
        result = Inpadoc.objects.get(publication="WO2009085664A2")
        assert len(result.family) >= 20

    def test_get_biblio_from_wo(self):
        result = Inpadoc.objects.get(publication="WO2009085664A2").biblio
        assert result.abstract is not None 


    def test_can_index_inpadoc_result(self):
        result = Inpadoc.objects.filter(applicant='Tesla')
        assert result[0] != result[1]


class TestInpadocBiblio():
    def test_inpadoc_biblio_manager(self):
        result = InpadocBiblio.objects.get(publication="USD870008")
        assert result.title == "Vehicle side skirt"
        assert result.number == 'D870008'
        assert result.country == 'US'


class TestInpadocLinks():
    def test_get_us_application_from_result(self):
        result = Inpadoc.objects.get(publication="US8131731B2")
        assert result.us_application.patent_title == "RELEVANCY SORTING OF USER'S BROWSER HISTORY"



import json
import logging
import math
from datetime import date
from datetime import datetime
from typing import Iterator
from typing import Union

import inflection
from patent_client import session
from patent_client.util.manager import Manager

from .model import USApplication
from .schema import USApplicationSchema

logger = logging.getLogger(__name__)


class HttpException(Exception):
    pass


class NotAvailableException(Exception):
    pass


QUERY_FIELDS = "appEarlyPubNumber applId appLocation appType appStatus_txt appConfrNumber appCustNumber appGrpArtNumber appCls appSubCls appEntityStatus_txt patentNumber patentTitle primaryInventor firstNamedApplicant appExamName appExamPrefrdName appAttrDockNumber appPCTNumber appIntlPubNumber wipoEarlyPubNumber pctAppType firstInventorFile appClsSubCls rankAndInventorsList"


class USApplicationManager(Manager[USApplication]):
    primary_key = "appl_id"
    query_url = "https://ped.uspto.gov/api/queries"
    page_size = 20
    __schema__ = USApplicationSchema()

    def __init__(self, *args, **kwargs):
        super(USApplicationManager, self).__init__(*args, **kwargs)
        self.pages = dict()

    def __len__(self):
        max_length = self.get_page(0)["numFound"] - self.config["offset"]
        limit = self.config["limit"]
        if not limit:
            return max_length
        else:
            return limit if limit < max_length else max_length

    def _get_results(self) -> Iterator[USApplication]:
        num_pages = math.ceil(len(self) / self.page_size)
        page_num = 0
        counter = 0
        while page_num < num_pages:
            page_data = self.get_page(page_num)
            for item in page_data["docs"]:
                if not self.config["limit"] or counter < self.config["limit"]:
                    yield self.__schema__.load(item)
                counter += 1
            page_num += 1

    def __iter__(self) -> Iterator[USApplication]:
        return super(USApplicationManager, self).__iter__()

    def get_page(self, page_number):
        if page_number not in self.pages:
            query_params = self.query_params(page_number)
            response = session.post(self.query_url, json=query_params, timeout=10)
            if not response.ok:
                if self.is_online():
                    raise HttpException(
                        f"{response.status_code}\n{response.text}\n{response.headers}\n{json.dumps(query_params)}"
                    )
            data = response.json()
            self.pages[page_number] = data["queryResults"]["searchResponse"]["response"]
        return self.pages[page_number]

    def query_params(self, page_no):
        sort_query = ""
        for s in self.config["order_by"]:
            if s[0] == "-":
                sort_query += f"{inflection.camelize(s[1:], uppercase_first_letter=False)} desc ".strip()
            else:
                sort_query += (
                    f"{inflection.camelize(s, uppercase_first_letter=False)} asc"
                ).strip()
        if not sort_query:
            sort_query = None

        query = list()
        mm_active = True
        for k, v in self.config["filter"].items():
            field = inflection.camelize(k, uppercase_first_letter=False)
            if not v:
                continue
            elif type(v) in (list, tuple):
                body = f" OR ".join(
                    f'"{value}"' if " " in value else value for value in v
                )
                mm_active = False
            else:
                body = v
            query.append(f"{field}:({body})")

        mm = "100%" if "appEarlyPubNumber" not in query else "90%"

        query = {
            "qf": QUERY_FIELDS,
            "fl": "*",  # ",".join(inflection.camelize(f, uppercase_first_letter=False) for f in RETURN_FIELDS),#"*",
            "searchText": " AND ".join(query).strip(),
            "sort": sort_query,
            "facet": "false",
            "mm": mm,
            "start": page_no * self.page_size + self.config["offset"],
            # "rows": self.page_size,
        }
        if not mm_active:
            del query["mm"]
        return query

    @property
    def allowed_filters(self):
        fields = self.fields()
        return list(fields.keys())

    def fields(self):
        if not hasattr(self.__class__, "_fields"):
            url = "https://ped.uspto.gov/api/search-fields"
            response = session.get(url)
            if not response.ok:
                raise ValueError("Can't get fields!")
            raw = response.json()
            output = {inflection.underscore(key): value for (key, value) in raw.items()}
            self.__class__._fields = output
        return self.__class__._fields

    def is_online(self):
        with session.cache_disabled():
            response = session.get("https://ped.uspto.gov/api/search-fields")
            if response.ok:
                return True
            elif "requested resource is not available" in response.text:
                raise NotAvailableException(
                    "Patent Examination Data is Offline - this is a USPTO problem"
                )
            elif "attempt failed or the origin closed the connection" in response.text:
                raise NotAvailableException(
                    "The Patent Examination Data API is Broken! - this is a USPTO problem"
                )
            else:
                raise NotAvailableException("There is a USPTO problem")

    @property
    def query_fields(self):
        fields = self.fields()
        for k in sorted(fields.keys()):
            if "facet" in k:
                continue
            print(f"{k} ({fields[k]})")


class DateEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, date):
            return o.isoformat()

        return json.JSONEncoder.default(self, o)



from marshmallow import fields
from .manager import ListManager 

class ListField(fields.List):
    def _deserialize(self, *args, **kwargs):
        out = super(ListField, self)._deserialize(*args, **kwargs)
        return ListManager(out)


from __future__ import annotations
import xml.etree.ElementTree as ET
import re
import json

from patent_client.util import Manager
from .session import session
from .schema import ITCInvestigationSchema, ITCDocumentSchema, ITCAttachmentSchema

BASE_URL = "https://edis.usitc.gov/data"

class ITCInvestigationManager(Manager['ITCInvestigation']):
    base_url = BASE_URL + "/investigation/"

    def _get_results(self):
        raise NotImplementedError("EDIS API does not have a search function!")

    def get(self, investigation_number):
        url = self.base_url + investigation_number
        response = session.get(url)
        return ITCInvestigationSchema().load(response.text)

class ITCDocumentManager(Manager['ITCDocument']):
    primary_key = "document_id"
    base_url = BASE_URL + "/document"
    allowed_filters = {
        "investigation_number": "investigationNumber",
        "phase": "investigationPhase",
        "type": "documentType",
        "firm": "firmOrg",
        "security": "securityLevel",
        "document_id": "document_id"
    }

    def get_document_by_id(self, document_id):
        breakpoint()
        response = session.get(
            f"{self.base_url}/{document_id}"
        )
        tree = ET.fromstring(response.text)
        doc_el = tree.find(".//document")
        return ITCDocumentSchema().load(doc_el)

    def _get_results(self):
        if "document_id" in self.config['filter']:
            yield self.get_document_by_id(self.config['document_id'])
        else:
            query = {self.allowed_filters[k]: v for (k, v) in self.config["filter"].items()}
            page = 1
            page_length = None
            while not page_length or page_length >= 100:
                query["pagenumber"] = page
                q_string = re.sub(
                    r'[\{\}":, ]+', "-", json.dumps(query, sort_keys=True)[1:-1]
                )
                response = session.get(
                    self.base_url, params=query
                )
                tree = ET.fromstring(response.text)[0]
                page_length = len(tree.findall("document"))
                for doc in tree.findall("document"):
                    yield ITCDocumentSchema().load(doc)
                page += 1

class ITCAttachmentManager(Manager['ITCAttachment']):
    primary_key = "id"
    base_url = BASE_URL + "/attachment/"
    allowed_filters = ["document_id"]

    def _get_results(self):
        doc_id = self.config["filter"]["document_id"]
        response = session.get(self.base_url + doc_id)
        tree = ET.fromstring(response.text)
        for element in tree.findall(".//attachment"):
            yield ITCAttachmentSchema().load(element)